# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import copy
import os

import libbe
import libbe.bug
import libbe.command
import libbe.diff
import libbe.command.util
import libbe.util.id
import libbe.util.tree


TAG="SUBSCRIBE:"


class Subscribe (libbe.command.Command):
    """(Un)subscribe to change notification

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Subscribe(ui=ui)

    >>> a = bd.bug_from_uuid('a')
    >>> print a.extra_strings
    []
    >>> ret = ui.run(cmd, {'subscriber':'John Doe <j@doe.com>'}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    Subscriptions for abc/a:
    John Doe <j@doe.com>    all    *
    >>> bd.flush_reload()
    >>> a = bd.bug_from_uuid('a')
    >>> print a.extra_strings
    ['SUBSCRIBE:John Doe <j@doe.com>\\tall\\t*']
    >>> ret = ui.run(cmd, {'subscriber':'Jane Doe <J@doe.com>', 'servers':'a.com,b.net'}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    Subscriptions for abc/a:
    Jane Doe <J@doe.com>    all    a.com,b.net
    John Doe <j@doe.com>    all    *
    >>> ret = ui.run(cmd, {'subscriber':'Jane Doe <J@doe.com>', 'servers':'a.edu'}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    Subscriptions for abc/a:
    Jane Doe <J@doe.com>    all    a.com,a.edu,b.net
    John Doe <j@doe.com>    all    *
    >>> ret = ui.run(cmd, {'unsubscribe':True, 'subscriber':'Jane Doe <J@doe.com>', 'servers':'a.com'}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    Subscriptions for abc/a:
    Jane Doe <J@doe.com>    all    a.edu,b.net
    John Doe <j@doe.com>    all    *
    >>> ret = ui.run(cmd, {'subscriber':'Jane Doe <J@doe.com>', 'servers':'*'}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    Subscriptions for abc/a:
    Jane Doe <J@doe.com>    all    *
    John Doe <j@doe.com>    all    *
    >>> ret = ui.run(cmd, {'unsubscribe':True, 'subscriber':'Jane Doe <J@doe.com>'}, ['/a']) # doctest: +NORMALIZE_WHITESPACE
    Subscriptions for abc/a:
    John Doe <j@doe.com>    all    *
    >>> ret = ui.run(cmd, {'unsubscribe':True, 'subscriber':'John Doe <j@doe.com>'}, ['/a'])
    >>> ret = ui.run(cmd,
    ...     {'subscriber':'Jane Doe <J@doe.com>', 'types':'new'},
    ...     [bd.uuid[:3]]) # doctest: +NORMALIZE_WHITESPACE
    Subscriptions for abc:
    Jane Doe <J@doe.com>    new    *
    >>> ret = ui.run(cmd,
    ...     {'subscriber':'Jane Doe <J@doe.com>'},
    ...     [bd.uuid]) # doctest: +NORMALIZE_WHITESPACE
    Subscriptions for abc:
    Jane Doe <J@doe.com>    all    *
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'subscribe'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='unsubscribe', short_name='u',
                    help='Unsubscribe instead of subscribing'),
                libbe.command.Option(name='list-all', short_name='a',
                    help='List all subscribers (no ID argument, read only action)'),
                libbe.command.Option(name='list', short_name='l',
                    help='List subscribers (read only action).'),
                libbe.command.Option(name='subscriber', short_name='s',
                    help='Email address of the subscriber (defaults to your user id).',
                    arg=libbe.command.Argument(
                        name='subscriber', metavar='EMAIL')),
                libbe.command.Option(name='servers', short_name='S',
                    help='Servers from which you want notification.',
                    arg=libbe.command.Argument(
                        name='servers', metavar='STRING')),
                libbe.command.Option(name='types', short_name='t',
                    help='Types of changes you wish to be notified about.',
                    arg=libbe.command.Argument(
                        name='types', metavar='STRING')),
                ])
        self.args.extend([
                libbe.command.Argument(
                    name='id', metavar='ID', default=tuple(),
                    optional=True, repeatable=True,
                    completion_callback=libbe.command.util.complete_bug_comment_id),
                ])

    def _run(self, **params):
        storage = self._get_storage()
        bugdirs = self._get_bugdirs()
        if params['list-all'] == True or params['list'] == True:
            writeable = storage.writeable
            storage.writeable = False
            if params['list-all'] == True:
                assert len(params['id']) == 0, params['id']
        subscriber = params['subscriber']
        if subscriber == None:
            subscriber = self._get_user_id()
        if params['unsubscribe'] == True:
            if params['servers'] == None:
                params['servers'] = 'INVALID'
            if params['types'] == None:
                params['types'] = 'INVALID'
        else:
            if params['servers'] == None:
                params['servers'] = '*'
            if params['types'] == None:
                params['types'] = 'all'
        servers = params['servers'].split(',')
        types = params['types'].split(',')

        if len(params['id']) == 0:
            params['id'] = bugdirs.keys()
        for _id in params['id']:
            p = libbe.util.id.parse_user(bugdirs, _id)
            if p['type'] == 'bugdir':
                type_root = libbe.diff.BUGDIR_TYPE_ALL
                entity = bugdirs[p['bugdir']]
            else: # bug-specific subscriptions
                type_root = libbe.diff.BUG_TYPE_ALL
                bugdir,bug,comment = (
                    libbe.command.util.bugdir_bug_comment_from_user_id(
                        bugdirs, _id))
                entity = bug
            entity_name = entity.id.user()
            if params['list-all'] == True:
                entity_name = 'anything in the bug directory'
            types = [libbe.diff.type_from_name(name, type_root, default=libbe.diff.INVALID_TYPE,
                                         default_ok=params['unsubscribe'])
                     for name in types]
            estrs = entity.extra_strings
            if params['list'] == True or params['list-all'] == True:
                pass
            else: # alter subscriptions
                if params['unsubscribe'] == True:
                    estrs = unsubscribe(estrs, subscriber, types, servers, type_root)
                else: # add the tag
                    estrs = subscribe(estrs, subscriber, types, servers, type_root)
                entity.extra_strings = estrs # reassign to notice change

            if params['list-all'] == True:
                subscriptions = []
                for bugdir in bugdirs.values():
                    bugdir.load_all_bugs()
                    subscriptions.extend(
                        get_bugdir_subscribers(bugdir, servers[0]))
            else:
                subscriptions = []
                for estr in entity.extra_strings:
                    if estr.startswith(TAG):
                        subscriptions.append(estr[len(TAG):])

            if len(subscriptions) > 0:
                print >> self.stdout, 'Subscriptions for %s:' % entity_name
                print >> self.stdout, '\n'.join(subscriptions)
        if params['list-all'] == True or params['list'] == True:
            storage.writeable = writeable
        return 0

    def _long_help(self):
        return """
ID can be either a bug ID, a bugdir ID, or blank, in which case it
refers to all known bugdirs.

SERVERS specifies the servers from which you would like to receive
notification.  Multiple severs may be specified in a comma-separated
list, or you can use "*" to match all servers (the default).  If you
have not selected a server, it should politely refrain from notifying
you of changes, although there is no way to guarantee this behavior.

Available TYPES:
  For bugs:
%s
  For %s:
%s

For unsubscription, any listed SERVERS and TYPES are removed from your
subscription.  Either the catch-all server "*" or type "%s" will
remove SUBSCRIBER entirely from the specified ID.

This command is intended for use primarily by public interfaces, since
if you're just hacking away on your private repository, you'll known
what's changed ;).  This command just (un)sets the appropriate
subscriptions, and leaves it up to each interface to perform the
notification.
""" % (libbe.diff.BUG_TYPE_ALL.string_tree(6), libbe.diff.BUGDIR_ID,
       libbe.diff.BUGDIR_TYPE_ALL.string_tree(6),
       libbe.diff.BUGDIR_TYPE_ALL)


# internal helper functions

def _generate_string(subscriber, types, servers):
    types = sorted([str(t) for t in types])
    servers = sorted(servers)
    return "%s%s\t%s\t%s" % (TAG,subscriber,",".join(types),",".join(servers))

def _parse_string(string, type_root):
    assert string.startswith(TAG), string
    string = string[len(TAG):]
    subscriber,types,servers = string.split("\t")
    types = [libbe.diff.type_from_name(name, type_root) for name in types.split(",")]
    return (subscriber,types,servers.split(","))

def _get_subscriber(extra_strings, subscriber, type_root):
    for i,string in enumerate(extra_strings):
        if string.startswith(TAG):
            s,ts,srvs = _parse_string(string, type_root)
            if s == subscriber:
                return i,s,ts,srvs # match!
    return None # no match

# functions exposed to other modules

def subscribe(extra_strings, subscriber, types, servers, type_root):
    args = _get_subscriber(extra_strings, subscriber, type_root)
    if args == None: # no match
        extra_strings.append(_generate_string(subscriber, types, servers))
        return extra_strings
    # Alter matched string
    i,s,ts,srvs = args
    for t in types:
        if t not in ts:
            ts.append(t)
    # remove descendant types
    all_ts = copy.copy(ts)
    for t in all_ts:
        for tt in all_ts:
            if tt in ts and t.has_descendant(tt):
                ts.remove(tt)
    if "*" in servers+srvs:
        srvs = ["*"]
    else:
        srvs = list(set(servers+srvs))
    extra_strings[i] = _generate_string(subscriber, ts, srvs)
    return extra_strings

def unsubscribe(extra_strings, subscriber, types, servers, type_root):
    args = _get_subscriber(extra_strings, subscriber, type_root)
    if args == None: # no match
        return extra_strings # pass
    # Remove matched string
    i,s,ts,srvs = args
    all_ts = copy.copy(ts)
    for t in types:
        for tt in all_ts:
            if tt in ts and t.has_descendant(tt):
                ts.remove(tt)
    if "*" in servers+srvs:
        srvs = []
    else:
        for srv in servers:
            if srv in srvs:
                srvs.remove(srv)
    if len(ts) == 0 or len(srvs) == 0:
        extra_strings.pop(i)
    else:
        extra_strings[i] = _generate_string(subscriber, ts, srvs)
    return extra_strings

def get_subscribers(extra_strings, type, server, type_root,
                    match_ancestor_types=False,
                    match_descendant_types=False):
    """
    Set match_ancestor_types=True if you want to find eveyone who
    cares about your particular type.

    Set match_descendant_types=True if you want to find subscribers
    who may only care about some subset of your type.  This is useful
    for generating lists of all the subscribers in a given set of
    extra_strings.

    >>> def sgs(*args, **kwargs):
    ...     return sorted(get_subscribers(*args, **kwargs))
    >>> es = []
    >>> es = subscribe(es, "John Doe <j@doe.com>", [libbe.diff.BUGDIR_TYPE_ALL],
    ...                ["a.com"], libbe.diff.BUGDIR_TYPE_ALL)
    >>> es = subscribe(es, "Jane Doe <J@doe.com>", [libbe.diff.BUGDIR_TYPE_NEW],
    ...                ["*"], libbe.diff.BUGDIR_TYPE_ALL)
    >>> sgs(es, libbe.diff.BUGDIR_TYPE_ALL, "a.com", libbe.diff.BUGDIR_TYPE_ALL)
    ['John Doe <j@doe.com>']
    >>> sgs(es, libbe.diff.BUGDIR_TYPE_ALL, "a.com", libbe.diff.BUGDIR_TYPE_ALL,
    ...     match_descendant_types=True)
    ['Jane Doe <J@doe.com>', 'John Doe <j@doe.com>']
    >>> sgs(es, libbe.diff.BUGDIR_TYPE_ALL, "b.net", libbe.diff.BUGDIR_TYPE_ALL,
    ...     match_descendant_types=True)
    ['Jane Doe <J@doe.com>']
    >>> sgs(es, libbe.diff.BUGDIR_TYPE_NEW, "a.com", libbe.diff.BUGDIR_TYPE_ALL)
    ['Jane Doe <J@doe.com>']
    >>> sgs(es, libbe.diff.BUGDIR_TYPE_NEW, "a.com", libbe.diff.BUGDIR_TYPE_ALL,
    ... match_ancestor_types=True)
    ['Jane Doe <J@doe.com>', 'John Doe <j@doe.com>']
    """
    for string in extra_strings:
        if not string.startswith(TAG):
            continue
        subscriber,types,servers = _parse_string(string, type_root)
        type_match = False
        if type in types:
            type_match = True
        if type_match == False and match_ancestor_types == True:
            for t in types:
                if t.has_descendant(type):
                    type_match = True
                    break
        if type_match == False and match_descendant_types == True:
            for t in types:
                if type.has_descendant(t):
                    type_match = True
                    break
        server_match = False
        if server in servers or servers == ["*"] or server == "*":
            server_match = True
        if type_match == True and server_match == True:
            yield subscriber

def get_bugdir_subscribers(bugdir, server):
    """
    I have a bugdir.  Who cares about it, and what do they care about?
    Returns a dict of dicts:
      subscribers[user][id] = types
    where id is either a bug.uuid (in the case of a bug subscription)
    or "%(bugdir_id)s" (in the case of a bugdir subscription).

    Only checks bugs that are currently in memory, so you might want
    to call bugdir.load_all_bugs() first.

    >>> bd = bugdir.SimpleBugDir(sync_with_disk=False)
    >>> a = bd.bug_from_shortname("a")
    >>> bd.extra_strings = subscribe(bd.extra_strings, "John Doe <j@doe.com>",
    ...                [libbe.diff.BUGDIR_TYPE_ALL], ["a.com"], libbe.diff.BUGDIR_TYPE_ALL)
    >>> bd.extra_strings = subscribe(bd.extra_strings, "Jane Doe <J@doe.com>",
    ...                [libbe.diff.BUGDIR_TYPE_NEW], ["*"], libbe.diff.BUGDIR_TYPE_ALL)
    >>> a.extra_strings = subscribe(a.extra_strings, "John Doe <j@doe.com>",
    ...                [libbe.diff.BUG_TYPE_ALL], ["a.com"], libbe.diff.BUG_TYPE_ALL)
    >>> subscribers = get_bugdir_subscribers(bd, "a.com")
    >>> subscribers["Jane Doe <J@doe.com>"]["%(bugdir_id)s"]
    [<SubscriptionType: new>]
    >>> subscribers["John Doe <j@doe.com>"]["%(bugdir_id)s"]
    [<SubscriptionType: all>]
    >>> subscribers["John Doe <j@doe.com>"]["a"]
    [<SubscriptionType: all>]
    >>> get_bugdir_subscribers(bd, "b.net")
    {'Jane Doe <J@doe.com>': {'%(bugdir_id)s': [<SubscriptionType: new>]}}
    >>> bd.cleanup()
    """ % {'bugdir_id':libbe.diff.BUGDIR_ID}
    subscribers = {}
    for sub in get_subscribers(bugdir.extra_strings, libbe.diff.BUGDIR_TYPE_ALL,
                               server, libbe.diff.BUGDIR_TYPE_ALL,
                               match_descendant_types=True):
        i,s,ts,srvs = _get_subscriber(bugdir.extra_strings, sub,
                                      libbe.diff.BUGDIR_TYPE_ALL)
        subscribers[sub] = {"DIR":ts}
    for bug in bugdir:
        for sub in get_subscribers(bug.extra_strings, libbe.diff.BUG_TYPE_ALL,
                                   server, libbe.diff.BUG_TYPE_ALL,
                                   match_descendant_types=True):
            i,s,ts,srvs = _get_subscriber(bug.extra_strings, sub,
                                          libbe.diff.BUG_TYPE_ALL)
            if sub in subscribers:
                subscribers[sub][bug.uuid] = ts
            else:
                subscribers[sub] = {bug.uuid:ts}
    return subscribers



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""
A command line interface to Bugs Everywhere.
"""

import optparse
import os
import sys

import libbe
import libbe.bugdir
import libbe.command
import libbe.command.help
import libbe.command.util
import libbe.storage
import libbe.version
import libbe.ui.util.pager
import libbe.util.encoding
import libbe.util.http


if libbe.TESTING == True:
    import doctest

class CallbackExit (Exception):
    pass

class CmdOptionParser(optparse.OptionParser):
    def __init__(self, command):
        self.command = command
        optparse.OptionParser.__init__(self)
        self.remove_option('-h')
        self.disable_interspersed_args()
        self._option_by_name = {}
        for option in self.command.options:
            self._add_option(option)
        self.set_usage(command.usage())


    def _add_option(self, option):
        option.validate()
        self._option_by_name[option.name] = option
        long_opt = '--%s' % option.name
        if option.short_name != None:
            short_opt = '-%s' % option.short_name
        assert '_' not in option.name, \
            'Non-reconstructable option name %s' % option.name
        kwargs = {'dest':option.name.replace('-', '_'),
                  'help':option.help}
        if option.arg == None: # a callback option
            kwargs['action'] = 'callback'
            kwargs['callback'] = self.callback
        elif option.arg.type == 'bool':
            kwargs['action'] = 'store_true'
            kwargs['metavar'] = None
            kwargs['default'] = False
        else:
            kwargs['type'] = option.arg.type
            kwargs['action'] = 'store'
            kwargs['metavar'] = option.arg.metavar
            kwargs['default'] = option.arg.default
        if option.short_name != None:
            opt = optparse.Option(short_opt, long_opt, **kwargs)
        else:
            opt = optparse.Option(long_opt, **kwargs)
        opt._option = option
        self.add_option(opt)

    def parse_args(self, args=None, values=None):
        args = self._get_args(args)
        options,parsed_args = optparse.OptionParser.parse_args(
            self, args=args, values=values)
        options = options.__dict__
        for name,value in options.items():
            if '_' in name: # reconstruct original option name
                options[name.replace('_', '-')] = options.pop(name)
        for name,value in options.items():
            argument = None
            option = self._option_by_name[name]
            if option.arg != None:
                argument = option.arg
            if value == '--complete':
                fragment = None
                indices = [i for i,arg in enumerate(args)
                           if arg == '--complete']
                for i in indices:
                    assert i > 0  # this --complete is an option value
                    if args[i-1] in ['--%s' % o.name
                                     for o in self.command.options]:
                        name = args[i-1][2:]
                        if name == option.name:
                            break
                    elif option.short_name != None \
                            and args[i-1].startswith('-') \
                            and args[i-1].endswith(option.short_name):
                        break
                if i+1 < len(args):
                    fragment = args[i+1]
                self.complete(argument, fragment)
            elif argument is not None:
                value = self.process_raw_argument(argument=argument, value=value)
                options[name] = value
        for i,arg in enumerate(parsed_args):
            if i > 0 and self.command.name == 'be':
                break # let this pass through for the command parser to handle
            elif i < len(self.command.args):
                argument = self.command.args[i]
            elif len(self.command.args) == 0:
                break # command doesn't take arguments
            else:
                argument = self.command.args[-1]
                if argument.repeatable == False:
                    raise libbe.command.UserError('Too many arguments')
            if arg == '--complete':
                fragment = None
                if i < len(parsed_args) - 1:
                    fragment = parsed_args[i+1]
                self.complete(argument, fragment)
            else:
                value = self.process_raw_argument(argument=argument, value=arg)
                parsed_args[i] = value
        if (len(parsed_args) > len(self.command.args) and
            (len(self.command.args) == 0 or
             self.command.args[-1].repeatable == False)):
            raise libbe.command.UserError('Too many arguments')
        for arg in self.command.args[len(parsed_args):]:
            if arg.optional == False:
                raise libbe.command.UsageError(
                    command=self.command,
                    message='Missing required argument %s' % arg.metavar)
        return (options, parsed_args)

    def callback(self, option, opt, value, parser):
        command_option = option._option
        if command_option.name == 'complete':
            argument = None
            fragment = None
            if len(parser.rargs) > 0:
                fragment = parser.rargs[0]
            self.complete(argument, fragment)
        else:
            print >> self.command.stdout, command_option.callback(
                self.command, command_option, value)
        raise CallbackExit

    def complete(self, argument=None, fragment=None):
        comps = self.command.complete(argument, fragment)
        if fragment != None:
            comps = [c for c in comps if c.startswith(fragment)]
        if len(comps) > 0:
            print >> self.command.stdout, '\n'.join(comps)
        raise CallbackExit

    def process_raw_argument(self, argument, value):
        if value == argument.default:
            return value
        if argument.type == 'string':
            if not hasattr(self, 'argv_encoding'):
                self.argv_encoding = libbe.util.encoding.get_argv_encoding()
            return unicode(value, self.argv_encoding)
        return value


class BE (libbe.command.Command):
    """Class for parsing the command line arguments for `be`.
    This class does not contain a useful _run() method.  Call this
    module's main() function instead.

    >>> ui = libbe.command.UserInterface()
    >>> ui.io.stdout = sys.stdout
    >>> be = BE(ui=ui)
    >>> ui.io.setup_command(be)
    >>> p = CmdOptionParser(be)
    >>> p.exit_after_callback = False
    >>> try:
    ...     options,args = p.parse_args(['--help']) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    ... except CallbackExit:
    ...     pass
    usage: be [options] [COMMAND [command-options] [COMMAND-ARGS ...]]
    <BLANKLINE>
    Options:
      -h, --help         Print a help message.
    <BLANKLINE>
      --complete         Print a list of possible completions.
    <BLANKLINE>
      --version          Print version string.
    ...
    >>> try:
    ...     options,args = p.parse_args(['--complete']) # doctest: +ELLIPSIS
    ... except CallbackExit:
    ...     print '  got callback'
    --help
    --version
    ...
    subscribe
    tag
    target
      got callback
    """
    name = 'be'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='version',
                    help='Print version string.',
                    callback=self.version),
                libbe.command.Option(name='full-version',
                    help='Print full version information.',
                    callback=self.full_version),
                libbe.command.Option(name='repo', short_name='r',
                    help='Select BE repository (see `be help repo`) rather '
                         'than the current directory.',
                    arg=libbe.command.Argument(
                        name='repo', metavar='REPO', default='.',
                        completion_callback=libbe.command.util.complete_path)),
                libbe.command.Option(name='server', short_name='s',
                    help='Select BE command server (see `be help '
                         'server`) rather than executing commands '
                         'locally',
                    arg=libbe.command.Argument(
                        name='server', metavar='URL')),
                libbe.command.Option(name='paginate',
                    help='Pipe all output into less (or if set, $PAGER).'),
                libbe.command.Option(name='no-pager',
                    help='Do not pipe git output into a pager.'),
                ])
        self.args.extend([
                libbe.command.Argument(
                    name='command', optional=False,
                    completion_callback=libbe.command.util.complete_command),
                libbe.command.Argument(
                    name='args', optional=True, repeatable=True)
                ])

    def usage(self):
        return 'usage: be [options] [COMMAND [command-options] [COMMAND-ARGS ...]]'

    def _long_help(self):
        cmdlist = []
        for name in libbe.command.commands():
            Class = libbe.command.get_command_class(command_name=name)
            assert hasattr(Class, '__doc__') and Class.__doc__ != None, \
                'Command class %s missing docstring' % Class
            cmdlist.append((Class.name, Class.__doc__.splitlines()[0]))
        cmdlist.sort()
        longest_cmd_len = max([len(name) for name,desc in cmdlist])
        ret = ['Bugs Everywhere - Distributed bug tracking',
               '', 'Commands:']
        for name, desc in cmdlist:
            numExtraSpaces = longest_cmd_len-len(name)
            ret.append('be {0}{1}    {2}'.format(
                    name, ' '*numExtraSpaces, desc))

        ret.extend(['', 'Topics:'])
        topic_list = [
            (name,desc.splitlines()[0])
            for name,desc in sorted(libbe.command.help.TOPICS.items())]
        longest_topic_len = max([len(name) for name,desc in topic_list])
        for name,desc in topic_list:
            extra_spaces = longest_topic_len - len(name)
            ret.append('{0}{1}    {2}'.format(name, ' '*extra_spaces, desc))

        ret.extend(['', 'Run', '  be help [command|topic]',
                    'for more information.'])
        return '\n'.join(ret)

    def version(self, *args):
        return libbe.version.version(verbose=False)

    def full_version(self, *args):
        return libbe.version.version(verbose=True)

class CommandLine (libbe.command.UserInterface):
    def __init__(self, *args, **kwargs):
        libbe.command.UserInterface.__init__(self, *args, **kwargs)
        self.restrict_file_access = False
        self.storage_callbacks = None
    def help(self):
        be = BE(ui=self)
        self.setup_command(be)
        return be.help()

def dispatch(ui, command, args):
    parser = CmdOptionParser(command)
    try:
        options,args = parser.parse_args(args)
        ret = ui.run(command, options, args)
    except CallbackExit:
        return 0
    except UnicodeDecodeError, e:
        print >> ui.io.stdout, '\n'.join([
                'ERROR:', str(e),
                'You should set a locale that supports unicode, e.g.',
                '  export LANG=en_US.utf8',
                'See http://docs.python.org/library/locale.html for details',
                ])
        return 1
    except libbe.command.UsageError, e:
        print >> ui.io.stdout, 'Usage Error:\n', e
        if e.command:
            print >> ui.io.stdout, e.command.usage()
        print >> ui.io.stdout, 'For usage information, try'
        print >> ui.io.stdout, '  be help %s' % e.command_name
        return 1
    except libbe.command.UserError, e:
        print >> ui.io.stdout, 'ERROR:\n', e
        return 1
    except OSError, e:
        print >> ui.io.stdout, 'OSError:\n', e
        return 1
    except libbe.storage.ConnectionError, e:
        print >> ui.io.stdout, 'Connection Error:\n', e
        return 1
    except libbe.util.http.HTTPError, e:
        print >> ui.io.stdout, 'HTTP Error:\n', e
        return 1
    except (libbe.util.id.MultipleIDMatches, libbe.util.id.NoIDMatches,
            libbe.util.id.InvalidIDStructure), e:
        print >> ui.io.stdout, 'Invalid id:\n', e
        return 1
    finally:
        command.cleanup()
    return ret

def main():
    io = libbe.command.StdInputOutput()
    ui = CommandLine(io)
    be = BE(ui=ui)
    ui.setup_command(be)

    parser = CmdOptionParser(be)
    try:
        options,args = parser.parse_args()
    except CallbackExit:
        return 0
    except libbe.command.UsageError, e:
        if isinstance(e.command, BE):
            # no command given, print usage string
            print >> ui.io.stdout, 'Usage Error:\n', e
            print >> ui.io.stdout, be.usage()
            print >> ui.io.stdout, 'For example, try'
            print >> ui.io.stdout, '  be help'
        else:
            print >> ui.io.stdout, 'Usage Error:\n', e
            if e.command:
                print >> ui.io.stdout, e.command.usage()
            print >> ui.io.stdout, 'For usage information, try'
            print >> ui.io.stdout, '  be help %s' % e.command_name
        return 1

    command_name = args.pop(0)
    try:
        Class = libbe.command.get_command_class(command_name=command_name)
    except libbe.command.UnknownCommand, e:
        print >> ui.io.stdout, e
        return 1

    ui.storage_callbacks = libbe.command.StorageCallbacks(options['repo'])
    command = Class(ui=ui, server=options['server'])
    ui.setup_command(command)

    if command.name in [
        'new', 'comment', 'commit', 'html', 'import-xml', 'serve-storage',
        'serve-commands']:
        paginate = 'never'
    else:
        paginate = 'auto'
    if options['paginate'] == True:
        paginate = 'always'
    if options['no-pager'] == True:
        paginate = 'never'
    libbe.ui.util.pager.run_pager(paginate)

    ret = dispatch(ui, command, args)
    try:
        ui.cleanup()
    except IOError, e:
        print >> ui.io.stdout, 'IOError:\n', e
        return 1

    return ret

if __name__ == '__main__':
    sys.exit(main())



# Copyright (C) 2008-2012 Ben Finney <benf@cybersource.com.au>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Robert Lehmann <mail@robertlehmann.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Git_ backend.

.. _Git: http://git-scm.com/
"""

import os
import os.path
import re
import shutil
import unittest

try:
    import pygit2 as _pygit2
except ImportError, error:
    _pygit2 = None
    _pygit2_import_error = error
else:
    if getattr(_pygit2, '__version__', '0.17.3') == '0.17.3':
        _pygit2 = None
        _pygit2_import_error = NotImplementedError(
            'pygit2 <= 0.17.3 not supported')

import libbe
from ...ui.util import user as _user
from ...util import encoding as _encoding
from ..base import EmptyCommit as _EmptyCommit
from . import base

if libbe.TESTING == True:
    import doctest
    import sys


def new():
    if _pygit2:
        return PygitGit()
    else:
        return ExecGit()


class PygitGit(base.VCS):
    """:py:class:`base.VCS` implementation for Git.

    Using :py:mod:`pygit2` for the Git activity.
    """
    name='pygit2'
    _null_hex = u'0' * 40
    _null_oid = '\00' * 20

    def __init__(self, *args, **kwargs):
        base.VCS.__init__(self, *args, **kwargs)
        self.versioned = True
        self._pygit_repository = None

    def __getstate__(self):
        """`pygit2.Repository`\s don't seem to pickle well.
        """
        attrs = dict(self.__dict__)
        if self._pygit_repository is not None:
            attrs['_pygit_repository'] = self._pygit_repository.path
        return attrs

    def __setstate__(self, state):
        """`pygit2.Repository`\s don't seem to pickle well.
        """
        self.__dict__.update(state)
        if self._pygit_repository is not None:
            gitdir = self._pygit_repository
            self._pygit_repository = _pygit2.Repository(gitdir)

    def _vcs_version(self):
        if _pygit2:
            return getattr(_pygit2, '__verison__', '?')
        return None

    def _vcs_get_user_id(self):
        try:
            name = self._pygit_repository.config['user.name']
        except KeyError:
            name = ''
        try:
            email = self._pygit_repository.config['user.email']
        except KeyError:
            email = ''
        if name != '' or email != '': # got something!
            # guess missing info, if necessary
            if name == '':
                name = _user.get_fallback_fullname()
            if email == '':
                email = _user.get_fallback_email()
            if '@' not in email:
                raise ValueError((name, email))
            return _user.create_user_id(name, email)
        return None # Git has no infomation

    def _vcs_detect(self, path):
        try:
            _pygit2.discover_repository(path)
        except KeyError:
            return False
        return True

    def _vcs_root(self, path):
        """Find the root of the deepest repository containing path."""
        # Assume that nothing funny is going on; in particular, that we aren't
        # dealing with a bare repo.
        gitdir = _pygit2.discover_repository(path)
        self._pygit_repository = _pygit2.Repository(gitdir)
        dirname,tip = os.path.split(gitdir)
        if tip == '':  # split('x/y/z/.git/') == ('x/y/z/.git', '')
            dirname,tip = os.path.split(dirname)
        assert tip == '.git', tip
        return dirname

    def _vcs_init(self, path):
        bare = False
        self._pygit_repository = _pygit2.init_repository(path, bare)

    def _vcs_destroy(self):
        vcs_dir = os.path.join(self.repo, '.git')
        if os.path.exists(vcs_dir):
            shutil.rmtree(vcs_dir)

    def _vcs_add(self, path):
        abspath = self._u_abspath(path)
        if os.path.isdir(abspath):
            return
        self._pygit_repository.index.read()
        self._pygit_repository.index.add(path)
        self._pygit_repository.index.write()

    def _vcs_remove(self, path):
        abspath = self._u_abspath(path)
        if not os.path.isdir(self._u_abspath(abspath)):
            self._pygit_repository.index.read()
            del self._pygit_repository.index[path]
            self._pygit_repository.index.write()
            os.remove(os.path.join(self.repo, path))

    def _vcs_update(self, path):
        self._vcs_add(path)

    def _git_get_commit(self, revision):
        if isinstance(revision, str):
            revision = unicode(revision, 'ascii')
        commit = self._pygit_repository.revparse_single(revision)
        assert commit.type == _pygit2.GIT_OBJ_COMMIT, commit
        return commit

    def _git_get_object(self, path, revision):
        commit = self._git_get_commit(revision=revision)
        tree = commit.tree
        sections = path.split(os.path.sep)
        for section in sections[:-1]:  # traverse trees
            child_tree = None
            for entry in tree:
                if entry.name == section:
                    eobj = entry.to_object()
                    if eobj.type == _pygit2.GIT_OBJ_TREE:
                        child_tree = eobj
                        break
                    else:
                        raise ValueError(path)  # not a directory
            if child_tree is None:
                raise ValueError((path, sections, section, [e.name for e in tree]))
                raise ValueError(path)  # not found
            tree = child_tree
        eobj = None
        for entry in tree:
            if entry.name == sections[-1]:
                eobj = entry.to_object()
        return eobj

    def _vcs_get_file_contents(self, path, revision=None):
        if revision == None:
            return base.VCS._vcs_get_file_contents(self, path, revision)
        else:
            blob = self._git_get_object(path=path, revision=revision)
            if blob.type != _pygit2.GIT_OBJ_BLOB:
                raise ValueError(path)  # not a file
            return blob.read_raw()

    def _vcs_path(self, id, revision):
        return self._u_find_id(id, revision)

    def _vcs_isdir(self, path, revision):
        obj = self._git_get_object(path=path, revision=revision)
        return obj.type == _pygit2.GIT_OBJ_TREE

    def _vcs_listdir(self, path, revision):
        tree = self._git_get_object(path=path, revision=revision)
        assert tree.type == _pygit2.GIT_OBJ_TREE, tree
        return [e.name for e in tree]

    def _vcs_commit(self, commitfile, allow_empty=False):
        self._pygit_repository.index.read()
        tree_oid = self._pygit_repository.index.write_tree()
        try:
            self._pygit_repository.head
        except _pygit2.GitError:  # no head; this is the first commit
            parents = []
            tree = self._pygit_repository[tree_oid]
            if not allow_empty and len(tree) == 0:
                raise _EmptyCommit()
        else:
            parents = [self._pygit_repository.head.oid]
            if (not allow_empty and
                tree_oid == self._pygit_repository.head.tree.oid):
                raise _EmptyCommit()
        update_ref = 'HEAD'
        user_id = self.get_user_id()
        name,email = _user.parse_user_id(user_id)
        # using default times is recent, see
        #   https://github.com/libgit2/pygit2/pull/129
        author = _pygit2.Signature(name, email)
        committer = author
        message = _encoding.get_file_contents(commitfile, decode=False)
        encoding = _encoding.get_text_file_encoding()
        commit_oid = self._pygit_repository.create_commit(
            update_ref, author, committer, message, tree_oid, parents,
            encoding)
        commit = self._pygit_repository[commit_oid]
        return commit.hex

    def _vcs_revision_id(self, index):
        walker = self._pygit_repository.walk(
            self._pygit_repository.head.oid, _pygit2.GIT_SORT_TIME)
        if index < 0:
            target_i = -1 - index  # -1: 0, -2: 1, ...
            for i,commit in enumerate(walker):
                if i == target_i:
                    return commit.hex
        elif index > 0:
            revisions = [commit.hex for commit in walker]
            # revisions is [newest, older, ..., oldest]
            if index > len(revisions):
                return None
            return revisions[len(revisions) - index]
        else:
            raise NotImplementedError('initial revision')
        return None

    def _vcs_changed(self, revision):
        commit = self._git_get_commit(revision=revision)
        diff = commit.tree.diff(self._pygit_repository.head.tree)
        new = set()
        modified = set()
        removed = set()
        for hunk in diff.changes['hunks']:
            if hunk.old_oid == self._null_hex:  # pygit2 uses hex in hunk.*_oid
                new.add(hunk.new_file)
            elif hunk.new_oid == self._null_hex:
                removed.add(hunk.old_file)
            else:
                modified.add(hunk.new_file)
        return (list(new), list(modified), list(removed))


class ExecGit (PygitGit):
    """:py:class:`base.VCS` implementation for Git.
    """
    name='git'
    client='git'

    def _vcs_version(self):
        try:
            status,output,error = self._u_invoke_client('--version')
        except CommandError:  # command not found?
            return None
        return output.strip()

    def _vcs_get_user_id(self):
        status,output,error = self._u_invoke_client(
            'config', 'user.name', expect=(0,1))
        if status == 0:
            name = output.rstrip('\n')
        else:
            name = ''
        status,output,error = self._u_invoke_client(
            'config', 'user.email', expect=(0,1))
        if status == 0:
            email = output.rstrip('\n')
        else:
            email = ''
        if name != '' or email != '': # got something!
            # guess missing info, if necessary
            if name == '':
                name = _user.get_fallback_fullname()
            if email == '':
                email = _user.get_fallback_email()
            return _user.create_user_id(name, email)
        return None # Git has no infomation

    def _vcs_detect(self, path):
        if self._u_search_parent_directories(path, '.git') != None :
            return True
        return False

    def _vcs_root(self, path):
        """Find the root of the deepest repository containing path."""
        # Assume that nothing funny is going on; in particular, that we aren't
        # dealing with a bare repo.
        if os.path.isdir(path) != True:
            path = os.path.dirname(path)
        status,output,error = self._u_invoke_client('rev-parse', '--git-dir',
                                                    cwd=path)
        gitdir = os.path.join(path, output.rstrip('\n'))
        dirname = os.path.abspath(os.path.dirname(gitdir))
        return dirname

    def _vcs_init(self, path):
        self._u_invoke_client('init', cwd=path)

    def _vcs_destroy(self):
        vcs_dir = os.path.join(self.repo, '.git')
        if os.path.exists(vcs_dir):
            shutil.rmtree(vcs_dir)

    def _vcs_add(self, path):
        if os.path.isdir(path):
            return
        self._u_invoke_client('add', path)

    def _vcs_remove(self, path):
        if not os.path.isdir(self._u_abspath(path)):
            self._u_invoke_client('rm', '-f', path)

    def _vcs_update(self, path):
        self._vcs_add(path)

    def _vcs_get_file_contents(self, path, revision=None):
        if revision == None:
            return base.VCS._vcs_get_file_contents(self, path, revision)
        else:
            arg = '%s:%s' % (revision,path)
            status,output,error = self._u_invoke_client('show', arg)
            return output

    def _vcs_path(self, id, revision):
        return self._u_find_id(id, revision)

    def _vcs_isdir(self, path, revision):
        arg = '%s:%s' % (revision,path)
        args = ['ls-tree', arg]
        kwargs = {'expect':(0,128)}
        status,output,error = self._u_invoke_client(*args, **kwargs)
        if status != 0:
            if 'not a tree object' in error:
                return False
            raise base.CommandError(args, status, stderr=error)
        return True

    def _vcs_listdir(self, path, revision):
        arg = '%s:%s' % (revision,path)
        status,output,error = self._u_invoke_client(
            'ls-tree', '--name-only', arg)
        return output.rstrip('\n').splitlines()

    def _vcs_commit(self, commitfile, allow_empty=False):
        args = ['commit', '--file', commitfile]
        if allow_empty == True:
            args.append('--allow-empty')
            status,output,error = self._u_invoke_client(*args)
        else:
            kwargs = {'expect':(0,1)}
            status,output,error = self._u_invoke_client(*args, **kwargs)
            strings = ['nothing to commit',
                       'nothing added to commit']
            if self._u_any_in_string(strings, output) == True:
                raise base.EmptyCommit()
        full_revision = self._vcs_revision_id(-1)
        assert full_revision[:7] in output, \
            'Mismatched revisions:\n%s\n%s' % (full_revision, output)
        return full_revision

    def _vcs_revision_id(self, index):
        args = ['rev-list', '--first-parent', '--reverse', 'HEAD']
        kwargs = {'expect':(0,128)}
        status,output,error = self._u_invoke_client(*args, **kwargs)
        if status == 128:
            if error.startswith("fatal: ambiguous argument 'HEAD': unknown "):
                return None
            raise base.CommandError(args, status, stderr=error)
        revisions = output.splitlines()
        try:
            if index > 0:
                return revisions[index-1]
            elif index < 0:
                return revisions[index]
            else:
                return None
        except IndexError:
            return None

    def _diff(self, revision):
        status,output,error = self._u_invoke_client('diff', revision)
        return output

    def _parse_diff(self, diff_text):
        """_parse_diff(diff_text) -> (new,modified,removed)

        `new`, `modified`, and `removed` are lists of files.

        Example diff text::

          diff --git a/dir/changed b/dir/changed
          index 6c3ea8c..2f2f7c7 100644
          --- a/dir/changed
          +++ b/dir/changed
          @@ -1,3 +1,3 @@
           hi
          -there
          +everyone and
           joe
          diff --git a/dir/deleted b/dir/deleted
          deleted file mode 100644
          index 225ec04..0000000
          --- a/dir/deleted
          +++ /dev/null
          @@ -1,3 +0,0 @@
          -in
          -the
          -beginning
          diff --git a/dir/moved b/dir/moved
          deleted file mode 100644
          index 5ef102f..0000000
          --- a/dir/moved
          +++ /dev/null
          @@ -1,4 +0,0 @@
          -the
          -ants
          -go
          -marching
          diff --git a/dir/moved2 b/dir/moved2
          new file mode 100644
          index 0000000..5ef102f
          --- /dev/null
          +++ b/dir/moved2
          @@ -0,0 +1,4 @@
          +the
          +ants
          +go
          +marching
          diff --git a/dir/new b/dir/new
          new file mode 100644
          index 0000000..94954ab
          --- /dev/null
          +++ b/dir/new
          @@ -0,0 +1,2 @@
          +hello
          +world
        """
        new = []
        modified = []
        removed = []
        lines = diff_text.splitlines()
        for i,line in enumerate(lines):
            if not line.startswith('diff '):
                continue
            file_a,file_b = line.split()[-2:]
            assert file_a.startswith('a/'), \
                'missformed file_a %s' % file_a
            assert file_b.startswith('b/'), \
                'missformed file_b %s' % file_b
            file = file_a[2:]
            assert file_b[2:] == file, \
                'diff file missmatch %s != %s' % (file_a, file_b)
            if lines[i+1].startswith('new '):
                new.append(file)
            elif lines[i+1].startswith('index '):
                modified.append(file)
            elif lines[i+1].startswith('deleted '):
                removed.append(file)
        return (new,modified,removed)

    def _vcs_changed(self, revision):
        return self._parse_diff(self._diff(revision))


if libbe.TESTING == True:
    base.make_vcs_testcase_subclasses(PygitGit, sys.modules[__name__])
    base.make_vcs_testcase_subclasses(ExecGit, sys.modules[__name__])

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Ben Finney <benf@cybersource.com.au>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Phil Schumm <philschumm@gmail.com>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Mercurial_ (hg) backend.

.. _Mercurial: http://mercurial.selenic.com/
"""

try:
    import mercurial
    import mercurial.dispatch
    import mercurial.ui
except ImportError:
    mercurial = None

try:
    # mercurial >= 1.2
    from mercurial.util import version
except ImportError:
    try:
        # mercurial <= 1.1.2
        from mercurial.version import get_version as version
    except ImportError:
        version = None

import os
import os.path
import re
import shutil
import StringIO
import sys
import time # work around http://mercurial.selenic.com/bts/issue618

import libbe
import base

if libbe.TESTING == True:
    import doctest
    import unittest


def new():
    return Hg()

class Hg(base.VCS):
    """:py:class:`base.VCS` implementation for Mercurial.
    """
    name='hg'
    client=None # mercurial module

    def __init__(self, *args, **kwargs):
        base.VCS.__init__(self, *args, **kwargs)
        self.versioned = True
        self.__updated = [] # work around http://mercurial.selenic.com/bts/issue618

    def _vcs_version(self):
        if version == None:
            return None
        return version()

    def _u_invoke_client(self, *args, **kwargs):
        if 'cwd' not in kwargs:
            kwargs['cwd'] = self.repo
        assert len(kwargs) == 1, kwargs
        fullargs = ['--cwd', kwargs['cwd']]
        fullargs.extend(args)
        cwd = os.getcwd()
        output = StringIO.StringIO()
        if self.version_cmp(1,9) >= 0:
            req = mercurial.dispatch.request(fullargs, fout=output)
            mercurial.dispatch.dispatch(req)
        else:
            stdout = sys.stdout
            sys.stdout = output
            mercurial.dispatch.dispatch(fullargs)
            sys.stdout = stdout
        os.chdir(cwd)
        return output.getvalue().rstrip('\n')

    def _vcs_get_user_id(self):
        output = self._u_invoke_client(
            'showconfig', 'ui.username').rstrip('\n')
        if output != '':
            return output
        return None

    def _vcs_detect(self, path):
        """Detect whether a directory is revision-controlled using Mercurial"""
        if self._u_search_parent_directories(path, '.hg') != None:
            return True
        return False

    def _vcs_root(self, path):
        return self._u_invoke_client('root', cwd=path)

    def _vcs_init(self, path):
        self._u_invoke_client('init', cwd=path)

    def _vcs_destroy(self):
        vcs_dir = os.path.join(self.repo, '.hg')
        if os.path.exists(vcs_dir):
            shutil.rmtree(vcs_dir)

    def _vcs_add(self, path):
        self._u_invoke_client('add', path)

    def _vcs_remove(self, path):
        self._u_invoke_client('rm', '--force', path)

    def _vcs_update(self, path):
        self.__updated.append(path) # work around http://mercurial.selenic.com/bts/issue618

    def _vcs_get_file_contents(self, path, revision=None):
        if revision == None:
            return base.VCS._vcs_get_file_contents(self, path, revision)
        else:
            return self._u_invoke_client('cat', '-r', revision, path)

    def _vcs_path(self, id, revision):
        manifest = self._u_invoke_client(
            'manifest', '--rev', revision).splitlines()
        return self._u_find_id_from_manifest(id, manifest, revision=revision)

    def _vcs_isdir(self, path, revision):
        output = self._u_invoke_client('manifest', '--rev', revision)
        files = output.splitlines()
        if path in files:
            return False
        return True

    def _vcs_listdir(self, path, revision):
        output = self._u_invoke_client('manifest', '--rev', revision)
        files = output.splitlines()
        path = path.rstrip(os.path.sep) + os.path.sep
        descendent_files = [self._u_rel_path(f, path) for f in files
                            if f.startswith(path)]
        return sorted(set(
                f.split(os.path.sep, 1)[0] for f in descendent_files))

    def _vcs_commit(self, commitfile, allow_empty=False):
        args = ['commit', '--logfile', commitfile]
        output = self._u_invoke_client(*args)
        # work around http://mercurial.selenic.com/bts/issue618
        strings = ['nothing changed']
        if self._u_any_in_string(strings, output) == True \
                and len(self.__updated) > 0:
            time.sleep(1)
            for path in self.__updated:
                os.utime(os.path.join(self.repo, path), None)
            output = self._u_invoke_client(*args)
        self.__updated = []
        # end work around
        if allow_empty == False:
            strings = ['nothing changed']
            if self._u_any_in_string(strings, output) == True:
                raise base.EmptyCommit()
        return self._vcs_revision_id(-1)

    def _vcs_revision_id(self, index, style='id'):
        if index > 0:
            index -= 1
        args = ['identify', '--rev', str(int(index)), '--%s' % style]
        output = self._u_invoke_client(*args)
        id = output.strip()
        if id == '000000000000':
            return None # before initial commit.
        return id

    def _diff(self, revision):
        return self._u_invoke_client(
            'diff', '-r', revision, '--git')

    def _parse_diff(self, diff_text):
        """_parse_diff(diff_text) -> (new,modified,removed)

        `new`, `modified`, and `removed` are lists of files.

        Example diff text::
                
          diff --git a/.be/dir/bugs/modified b/.be/dir/bugs/modified
          --- a/.be/dir/bugs/modified
          +++ b/.be/dir/bugs/modified
          @@ -1,1 +1,1 @@ some value to be modified
          -some value to be modified
          \ No newline at end of file
          +a new value
          \ No newline at end of file
          diff --git a/.be/dir/bugs/moved b/.be/dir/bugs/moved
          deleted file mode 100644
          --- a/.be/dir/bugs/moved
          +++ /dev/null
          @@ -1,1 +0,0 @@
          -this entry will be moved
          \ No newline at end of file
          diff --git a/.be/dir/bugs/moved2 b/.be/dir/bugs/moved2
          new file mode 100644
          --- /dev/null
          +++ b/.be/dir/bugs/moved2
          @@ -0,0 +1,1 @@
          +this entry will be moved
          \ No newline at end of file
          diff --git a/.be/dir/bugs/new b/.be/dir/bugs/new
          new file mode 100644
          --- /dev/null
          +++ b/.be/dir/bugs/new
          @@ -0,0 +1,1 @@
          +this entry is new
          \ No newline at end of file
          diff --git a/.be/dir/bugs/removed b/.be/dir/bugs/removed
          deleted file mode 100644
          --- a/.be/dir/bugs/removed
          +++ /dev/null
          @@ -1,1 +0,0 @@
          -this entry will be deleted
          \ No newline at end of file
        """
        new = []
        modified = []
        removed = []
        lines = diff_text.splitlines()
        for i,line in enumerate(lines):
            if not line.startswith('diff '):
                continue
            file_a,file_b = line.split()[-2:]
            assert file_a.startswith('a/'), \
                'missformed file_a %s' % file_a
            assert file_b.startswith('b/'), \
                'missformed file_b %s' % file_b
            file = file_a[2:]
            assert file_b[2:] == file, \
                'diff file missmatch %s != %s' % (file_a, file_b)
            if lines[i+1].startswith('new '):
                new.append(file)
            elif lines[i+1].startswith('deleted '):
                removed.append(file)
            else:
                modified.append(file)
        return (new,modified,removed)

    def _vcs_changed(self, revision):
        return self._parse_diff(self._diff(revision))


if libbe.TESTING == True:
    base.make_vcs_testcase_subclasses(Hg, sys.modules[__name__])

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2010-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define an HTTP-based :py:class:`~libbe.storage.base.VersionedStorage`
implementation.

See Also
--------
:py:mod:`libbe.command.serve_storage` : the associated server
"""

from __future__ import absolute_import
import sys
import urllib
import urlparse

import libbe
import libbe.version
import libbe.util.http
from libbe.util.http import HTTP_VALID, HTTP_USER_ERROR
from . import base

from libbe import TESTING

if TESTING == True:
    import copy
    import doctest
    import StringIO
    import unittest

    import libbe.bugdir
    import libbe.command.serve_storage
    import libbe.util.http
    import libbe.util.wsgi


class HTTP (base.VersionedStorage):
    """:py:class:`~libbe.storage.base.VersionedStorage` implementation over
    HTTP.

    Uses GET to retrieve information and POST to set information.
    """
    name = 'HTTP'
    user_agent = 'BE-HTTP-Storage'

    def __init__(self, repo, *args, **kwargs):
        repo,self.uname,self.password = self.parse_repo(repo)
        base.VersionedStorage.__init__(self, repo, *args, **kwargs)

    def parse_repo(self, repo):
        """Grab username and password (if any) from the repo URL.

        Examples
        --------

        >>> s = HTTP('http://host.com/path/to/repo')
        >>> s.repo
        'http://host.com/path/to/repo'
        >>> s.uname == None
        True
        >>> s.password == None
        True
        >>> s.parse_repo('http://joe:secret@host.com/path/to/repo')
        ('http://host.com/path/to/repo', 'joe', 'secret')
        """
        scheme,netloc,path,params,query,fragment = urlparse.urlparse(repo)
        parts = netloc.split('@', 1)
        if len(parts) == 2:
            uname,password = parts[0].split(':')
            repo = urlparse.urlunparse(
                (scheme, parts[1], path, params, query, fragment))
        else:
            uname,password = (None, None)
        return (repo, uname, password)

    def get_post_url(self, url, get=True, data_dict=None, headers=[]):
        if self.uname != None and self.password != None:
            headers.append(('Authorization','Basic %s' % \
                ('%s:%s' % (self.uname, self.password)).encode('base64')))
        return libbe.util.http.get_post_url(
            url, get, data_dict=data_dict, headers=headers,
            agent=self.user_agent)

    def storage_version(self, revision=None):
        """Return the storage format for this backend."""
        return libbe.storage.STORAGE_VERSION

    def _init(self):
        """Create a new storage repository."""
        raise base.NotSupported(
            'init', 'Cannot initialize this repository format.')

    def _destroy(self):
        """Remove the storage repository."""
        raise base.NotSupported(
            'destroy', 'Cannot destroy this repository format.')

    def _connect(self):
        self.check_storage_version()

    def _disconnect(self):
        pass

    def _add(self, id, parent=None, directory=False):
        url = urlparse.urljoin(self.repo, 'add')
        page,final_url,info = self.get_post_url(
            url, get=False,
            data_dict={'id':id, 'parent':parent, 'directory':directory})

    def _exists(self, id, revision=None):
        url = urlparse.urljoin(self.repo, 'exists')
        page,final_url,info = self.get_post_url(
            url, get=True,
            data_dict={'id':id, 'revision':revision})
        if page == 'True':
            return True
        return False

    def _remove(self, id):
        url = urlparse.urljoin(self.repo, 'remove')
        page,final_url,info = self.get_post_url(
            url, get=False,
            data_dict={'id':id, 'recursive':False})

    def _recursive_remove(self, id):
        url = urlparse.urljoin(self.repo, 'remove')
        page,final_url,info = self.get_post_url(
            url, get=False,
            data_dict={'id':id, 'recursive':True})

    def _ancestors(self, id=None, revision=None):
        url = urlparse.urljoin(self.repo, 'ancestors')
        page,final_url,info = self.get_post_url(
            url, get=True,
            data_dict={'id':id, 'revision':revision})
        return page.strip('\n').splitlines()

    def _children(self, id=None, revision=None):
        url = urlparse.urljoin(self.repo, 'children')
        page,final_url,info = self.get_post_url(
            url, get=True,
            data_dict={'id':id, 'revision':revision})
        return page.strip('\n').splitlines()

    def _get(self, id, default=base.InvalidObject, revision=None):
        url = urlparse.urljoin(self.repo, '/'.join(['get', id]))
        try:
            page,final_url,info = self.get_post_url(
                url, get=True,
                data_dict={'revision':revision})
        except libbe.util.http.HTTPError, e:
            if not (hasattr(e.error, 'code') and e.error.code in HTTP_VALID):
                raise
            elif default == base.InvalidObject:
                raise base.InvalidID(id)
            return default
        version = info['X-BE-Version']
        if version != libbe.storage.STORAGE_VERSION:
            raise base.InvalidStorageVersion(
                version, libbe.storage.STORAGE_VERSION)
        return page

    def _set(self, id, value):
        url = urlparse.urljoin(self.repo, '/'.join(['set', id]))
        try:
            page,final_url,info = self.get_post_url(
                url, get=False,
                data_dict={'value':value})
        except libbe.util.http.HTTPError, e:
            if not (hasattr(e.error, 'code') and e.error.code in HTTP_VALID):
                raise
            if e.error.code == HTTP_USER_ERROR \
                    and not 'InvalidID' in str(e.error):
                raise base.InvalidDirectory(
                    'Directory %s cannot have data' % id)
            raise base.InvalidID(id)

    def _commit(self, summary, body=None, allow_empty=False):
        url = urlparse.urljoin(self.repo, 'commit')
        try:
            page,final_url,info = self.get_post_url(
                url, get=False,
                data_dict={'summary':summary, 'body':body,
                           'allow_empty':allow_empty})
        except libbe.util.http.HTTPError, e:
            if not (hasattr(e.error, 'code') and e.error.code in HTTP_VALID):
                raise
            if e.error.code == HTTP_USER_ERROR:
                raise base.EmptyCommit
            raise base.InvalidID(id)
        return page.rstrip('\n')

    def revision_id(self, index=None):
        """Return the name of the <index>th revision.

        The choice of which branch to follow when crossing
        branches/merges is not defined.  Revision indices start at 1;
        ID 0 is the blank repository.

        Return None if index==None.

        Raises
        ------
        InvalidRevision
          If the specified revision does not exist.
        """
        if index == None:
            return None
        try:
            if int(index) != index:
                raise base.InvalidRevision(index)
        except ValueError:
            raise base.InvalidRevision(index)
        url = urlparse.urljoin(self.repo, 'revision-id')
        try:
            page,final_url,info = self.get_post_url(
                url, get=True,
                data_dict={'index':index})
        except libbe.util.http.HTTPError, e:
            if not (hasattr(e.error, 'code') and e.error.code in HTTP_VALID):
                raise
            if e.error.code == HTTP_USER_ERROR:
                raise base.InvalidRevision(index)
            raise base.InvalidID(id)
        return page.rstrip('\n')

    def changed(self, revision=None):
        url = urlparse.urljoin(self.repo, 'changed')
        page,final_url,info = self.get_post_url(
            url, get=True,
            data_dict={'revision':revision})
        lines = page.strip('\n')
        new,mod,rem = [p.splitlines() for p in page.split('\n\n')]
        return (new, mod, rem)

    def check_storage_version(self):
        version = self.storage_version()
        if version != libbe.storage.STORAGE_VERSION:
            raise base.InvalidStorageVersion(
                version, libbe.storage.STORAGE_VERSION)

    def storage_version(self, revision=None):
        url = urlparse.urljoin(self.repo, 'version')
        page,final_url,info = self.get_post_url(
            url, get=True, data_dict={'revision':revision})
        return page.rstrip('\n')

if TESTING == True:
    class TestingHTTP (HTTP):
        name = 'TestingHTTP'
        def __init__(self, repo, *args, **kwargs):
            self._storage_backend = base.VersionedStorage(repo)
            app = libbe.command.serve_storage.ServerApp(
                storage=self._storage_backend)
            self.app = libbe.util.wsgi.BEExceptionApp(app=app)
            HTTP.__init__(self, repo='http://localhost:8000/', *args, **kwargs)
            self.intitialized = False
            # duplicated from libbe.util.wsgi.WSGITestCase
            self.default_environ = {
                'REQUEST_METHOD': 'GET', # 'POST', 'HEAD'
                'REMOTE_ADDR': '192.168.0.123',
                'SCRIPT_NAME':'',
                'PATH_INFO': '',
                #'QUERY_STRING':'',   # may be empty or absent
                #'CONTENT_TYPE':'',   # may be empty or absent
                #'CONTENT_LENGTH':'', # may be empty or absent
                'SERVER_NAME':'example.com',
                'SERVER_PORT':'80',
                'SERVER_PROTOCOL':'HTTP/1.1',
                'wsgi.version':(1,0),
                'wsgi.url_scheme':'http',
                'wsgi.input':StringIO.StringIO(),
                'wsgi.errors':StringIO.StringIO(),
                'wsgi.multithread':False,
                'wsgi.multiprocess':False,
                'wsgi.run_once':False,
                }
        def getURL(self, app, path='/', method='GET', data=None,
                   scheme='http', environ={}):
            # duplicated from libbe.util.wsgi.WSGITestCase
            env = copy.copy(self.default_environ)
            env['PATH_INFO'] = path
            env['REQUEST_METHOD'] = method
            env['scheme'] = scheme
            if data != None:
                enc_data = urllib.urlencode(data)
                if method == 'POST':
                    env['CONTENT_LENGTH'] = len(enc_data)
                    env['wsgi.input'] = StringIO.StringIO(enc_data)
                else:
                    assert method in ['GET', 'HEAD'], method
                    env['QUERY_STRING'] = enc_data
            for key,value in environ.items():
                env[key] = value
            try:
                result = app(env, self.start_response)
            except libbe.util.wsgi.HandlerError as e:
                raise libbe.util.http.HTTPError(error=e, url=path, msg=str(e))
            return ''.join(result)
        def start_response(self, status, response_headers, exc_info=None):
            self.status = status
            self.response_headers = response_headers
            self.exc_info = exc_info
        def get_post_url(self, url, get=True, data_dict=None, headers=[]):
            if get == True:
                method = 'GET'
            else:
                method = 'POST'
            scheme,netloc,path,params,query,fragment = urlparse.urlparse(url)
            environ = {}
            for header_name,header_value in headers:
                environ['HTTP_%s' % header_name] = header_value
            output = self.getURL(
                self.app, path, method, data_dict, scheme, environ)
            if self.status != '200 OK':
                class __estr (object):
                    def __init__(self, string):
                        self.string = string
                        self.code = int(string.split()[0])
                    def __str__(self):
                        return self.string
                error = __estr(self.status)
                raise libbe.util.http.HTTPError(
                    error=error, url=url, msg=output)
            info = dict(self.response_headers)
            return (output, url, info)
        def _init(self):
            try:
                HTTP._init(self)
                raise AssertionError
            except base.NotSupported:
                pass
            self._storage_backend._init()
        def _destroy(self):
            try:
                HTTP._destroy(self)
                raise AssertionError
            except base.NotSupported:
                pass
            self._storage_backend._destroy()
        def _connect(self):
            self._storage_backend._connect()
            HTTP._connect(self)
        def _disconnect(self):
            HTTP._disconnect(self)
            self._storage_backend._disconnect()


    base.make_versioned_storage_testcase_subclasses(
        TestingHTTP, sys.modules[__name__])

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])
